/********************************************************/
/*  ntapi: Native API core library                      */
/*  Copyright (C) 2013--2017  Z. Gilboa                 */
/*  Released under GPLv2 and GPLv3; see COPYING.NTAPI.  */
/********************************************************/

#include <psxtypes/psxtypes.h>
#include <ntapi/nt_file.h>
#include <ntapi/nt_string.h>
#include <ntapi/nt_atomic.h>
#include <ntapi/nt_port.h>
#include <ntapi/nt_ipc.h>
#include <ntapi/nt_msq.h>
#include <ntapi/ntapi.h>
#include "ntapi_impl.h"

int32_t	__stdcall __ntapi_msq_set(
	__in	nt_msq_info *		msq,
	__out	nt_io_status_block *	iosb,
	__in	void *			msq_info,
	__in	uint32_t		msq_info_length,
	__in	int32_t			msq_info_class)
{
	int32_t			status;
	nt_msq_info_msg		msg;

	/* validate */
	if (!iosb)
		return NT_STATUS_INVALID_PARAMETER;

	else if (!msq_info)
		return NT_STATUS_INVALID_PARAMETER;

	else if (msq_info_length != sizeof(msg.data.msqinfo))
		return NT_STATUS_INFO_LENGTH_MISMATCH;

	else if (msq_info_class)
		return NT_STATUS_INVALID_INFO_CLASS;

	/* msg */
	__ntapi->tt_aligned_block_memset(
		&msg,0,sizeof(msg));

	msg.header.msg_type		= NT_LPC_NEW_MESSAGE;
	msg.header.data_size		= sizeof(msg.data);
	msg.header.msg_size		= sizeof(msg);
	msg.data.ttyinfo.opcode		= NT_TTY_MSQ_SET;

	__ntapi->tt_generic_memcpy(
		&msg.data.msqinfo,
		msq_info,
		sizeof(msg.data.msqinfo));

	msg.data.msqinfo.msqkey		= msq->msqkey;
	msg.data.msqinfo.msqid		= msq->msqid;
	msg.data.msqinfo.section_size	= msq_info_length;

	if ((status = __ntapi->zw_request_wait_reply_port(msq->hport,&msg,&msg)))
		return status;
	else if (msg.data.ttyinfo.status)
		return msg.data.ttyinfo.status;

	/* reply */
	iosb->status = NT_STATUS_SUCCESS;
	iosb->info   = 0;

	return NT_STATUS_SUCCESS;
}
